/*

mzlibsql.cpp

Code based on Example#3 in mini zlib sample code.
code.google.com/p/miniz/

File miniz.c should be placed in the same IFS folder as mzlibsql

Compile Instructions:
CRTCPPMOD MODULE(yyyyy/mzlibsql) SRCSTMF('/source/mzlibsql.c')
CRTSRVPGM SRVPGM(yyyyy/mzlibsql) EXPORT(*ALL)

DB2 User Defined Function Registration:
CREATE OR REPLACE FUNCTION yyyyy.DEFLATE(DATA BLOB(2G) AS LOCATOR)
RETURNS BLOB(2G) AS LOCATOR
SPECIFIC DEFLATE_BLOB
EXTERNAL NAME 'yyyyy/MZLIBSQL(zlib_deflate)'
LANGUAGE C++
DETERMINISTIC
NO SQL
NOT FENCED
PARAMETER STYLE SQL
RETURNS NULL ON NULL INPUT
NO EXTERNAL ACTION
;
CREATE OR REPLACE FUNCTION yyyyy.DEFLATE(DATA CLOB(2G) AS LOCATOR)
RETURNS BLOB(2G) AS LOCATOR
SPECIFIC DEFLATE_CLOB
EXTERNAL NAME 'yyyyy/MZLIBSQL(zlib_deflate)'
LANGUAGE C++
DETERMINISTIC
NO SQL
NOT FENCED
PARAMETER STYLE SQL
RETURNS NULL ON NULL INPUT
NO EXTERNAL ACTION
;
CREATE OR REPLACE FUNCTION yyyyy.DEFLATE(DATA XML AS LOCATOR)
RETURNS BLOB(2G) AS LOCATOR
SPECIFIC DEFLATE_XML
EXTERNAL NAME 'yyyyy/MZLIBSQL(zlib_deflate)'
LANGUAGE C++
DETERMINISTIC
NO SQL
NOT FENCED
PARAMETER STYLE SQL
RETURNS NULL ON NULL INPUT
NO EXTERNAL ACTION
;
CREATE OR REPLACE FUNCTION yyyyy.INFLATE(DATA BLOB(2G) AS LOCATOR)
RETURNS BLOB(2G) AS LOCATOR
SPECIFIC INFLATE_BLOB
EXTERNAL NAME 'yyyyy/MZLIBSQL(zlib_inflate)'
LANGUAGE C++
DETERMINISTIC
NO SQL
NOT FENCED
PARAMETER STYLE SQL
RETURNS NULL ON NULL INPUT
NO EXTERNAL ACTION
;
CREATE OR REPLACE FUNCTION yyyyy.INFLATE_CLOB(DATA BLOB(2G) AS LOCATOR)
RETURNS CLOB(2G) AS LOCATOR
SPECIFIC INFLATE_CLOB
EXTERNAL NAME 'yyyyy/MZLIBSQL(zlib_inflate)'
LANGUAGE C++
DETERMINISTIC
NO SQL
NOT FENCED
PARAMETER STYLE SQL
RETURNS NULL ON NULL INPUT
NO EXTERNAL ACTION
;
CREATE OR REPLACE FUNCTION yyyyy.INFLATE_XML(DATA BLOB(2G) AS LOCATOR)
RETURNS XML AS LOCATOR
SPECIFIC INFLATE_XML
EXTERNAL NAME 'yyyyy/MZLIBSQL(zlib_inflate)'
LANGUAGE C++
DETERMINISTIC
NO SQL
NOT FENCED
PARAMETER STYLE SQL
RETURNS NULL ON NULL INPUT
NO EXTERNAL ACTION
;
// Test
set path=yyyyy;
SELECT INFLATE_CLOB(DEFLATE(CAST('Mike Mike Mike Mike Mike Mike Mike Mike ' AS BLOB(1K))))
FROM QSYS2.QSQPTABL;
;
SELECT LENGTH(INFLATE_CHAR(DEFLATE(BLOB(REPEAT('axy',2500))||BLOB(SPACE(32600))||BLOB(SPACE(32600))||BLOB(SPACE(32600))||BLOB(SPACE(32600))||BLOB(SPACE(32600))||BLOB(SPACE(32600))||BLOB(SPACE(32600))||CAST('Mike Mike Mike Mike Mike Mike Mikexxx Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike Mike ' AS BLOB(1M)))))
FROM QSYS2.QSQPTABL;

*/
#include "stdio.h"
#include "string.h"
#include "stdlib.h"

// Mini zlib
#include "miniz.c"

// DB2 includes
#include "sql.h"
#include "sqludf.h"

#define my_max(a,b) (((a) > (b)) ? (a) : (b))
#define my_min(a,b) (((a) < (b)) ? (a) : (b))
typedef unsigned char uint8;
typedef unsigned short uint16;
typedef unsigned int uint;

int rc;
// int level = Z_BEST_COMPRESSION;
int level = 6;

// Buffer size is 512K
#define BUF_SIZE (1024 * 512)

#define mcompress zip_compress
#define MINIZ_NO_ARCHIVE_APIS
#define MINIZ_NO_STDIO
#define MINIZ_NO_ARCHIVE_WRITING_APIS

extern "C" {

	void zlib_deflate
	(udf_locator *input,  /* ptr to input  argument */
	 udf_locator *output, /* ptr to output argument */
	short *input_ind,  /* ptr to input indicator  */
	short *output_ind, /* ptr to output indicator */
	char sqlstate[6], /* sqlstate */
	char fname[140], /* fully qualified function name */
	char finst[129], /* function specific name */
	char msgtext[71]) /* msg text buffer */
	{

		*output_ind=*input_ind;
		if (*input_ind!=0)
			return;

		uint inlob_size;
		z_stream stream;

		sqlint32 length;
		rc=inlob_size=sqludf_length(input, &length);
		inlob_size=(uint)length;

		static uint8 s_inbuf[BUF_SIZE];
		static uint8 s_outbuf[BUF_SIZE];

		// Init the z_stream
		memset(&stream, 0, sizeof(stream));
		stream.next_in = s_inbuf;
		stream.avail_in = 0;
		stream.next_out = s_outbuf;
		stream.avail_out = BUF_SIZE;

		// Create output BLOB locator
		rc=sqludf_create_locator(SQL_TYP_BLOB, &output);
		if (rc!=0) {
			strcpy(sqlstate,"38U01");
			strcpy(msgtext,"create out LOB locator hosed\0");
			return;
		}

		// Compression.
		uint inlob_remaining = inlob_size;
		if (deflateInit(&stream, level) != Z_OK)
		{
			strcpy(sqlstate,"38U02");
			strcpy(msgtext,"deflateInit() failed");
			return;
		}

		uint input_pos=1;
		for ( ; ; )
		{
			int status;
			if (!stream.avail_in)
			{
				uint n = my_min(BUF_SIZE, inlob_remaining);
				// Input buffer is empty, so read more bytes from input lob        
				rc = sqludf_substr(input, input_pos, n, s_inbuf,&length);
				if (length != n)
				{
					strcpy(sqlstate,"38U03");
					strcpy(msgtext,"Failed reading from input lob");
					return;
				}

				stream.next_in = s_inbuf;
				stream.avail_in = n;

				inlob_remaining -= n;
				input_pos += n;
			}

			status = deflate(&stream, inlob_remaining ? Z_NO_FLUSH : Z_FINISH);

			if ((status == Z_STREAM_END) || (!stream.avail_out))
			{
				// Output buffer is full, or compression is done, so write buffer to output file.
				uint n = BUF_SIZE - stream.avail_out;
				rc = sqludf_append(output,
								  (unsigned char *) s_outbuf, n, &length);

				if (rc || length != n)
				{
					strcpy(sqlstate,"38U04");
					strcpy(msgtext,"Failed writing to output lob");
					return;
				}
				stream.next_out = s_outbuf;
				stream.avail_out = BUF_SIZE;
			}

			if (status == Z_STREAM_END)
				break;
			else if (status != Z_OK)
			{
				strcpy(sqlstate,"38U05");
				strcpy(msgtext,"deflate() failed");
				return;
			}
		}

		if (deflateEnd(&stream) != Z_OK)
		{
			strcpy(sqlstate,"38U06");
			strcpy(msgtext,"deflateEnd() failed");
			return;
		}

		return;
	}

	void zlib_inflate
	(udf_locator *input,  /* ptr to input  argument */
	 udf_locator *output, /* ptr to output argument */
	short *input_ind,  /* ptr to input indicator  */
	short *output_ind, /* ptr to output indicator */
	char sqlstate[6], /* sqlstate */
	char fname[140], /* fully qualified function name */
	char finst[129], /* function specific name */
	char msgtext[71]) /* msg text buffer */
	{
		// Create output BLOB locator
		rc=sqludf_create_locator(SQL_TYP_BLOB, &output);
		if (rc!=0) {
			strcpy(sqlstate,"38U01");
			strcpy(msgtext,"create out LOB locator hosed\0");
			return;
		}

		*output_ind=*input_ind;
		if (*input_ind)
			return;

		uint inlob_size;
		z_stream stream;

		sqlint32 length;
		rc=inlob_size=sqludf_length(input, &length);
		inlob_size=(uint)length;

		static uint8 s_inbuf[BUF_SIZE];
		static uint8 s_outbuf[BUF_SIZE];

		// Init the z_stream
		memset(&stream, 0, sizeof(stream));
		stream.next_in = s_inbuf;
		stream.avail_in = 0;
		stream.next_out = s_outbuf;
		stream.avail_out = BUF_SIZE;

		// Decompression.
		uint inlob_remaining = inlob_size;

		if (inflateInit(&stream))
		{
			strcpy(sqlstate,"38U02");
			strcpy(msgtext,"inflate init failed");
			return;
		}

		uint input_pos=1;
		for ( ; ; )
		{
			int status;
			if (!stream.avail_in)
			{
				// Input buffer is empty, so read more bytes from input LOB
				uint n = my_min(BUF_SIZE, inlob_remaining);
				rc = sqludf_substr(input, input_pos, n, s_inbuf, &length);
				if (rc || length != n)
				{
					strcpy(sqlstate,"38U03");
					strcpy(msgtext,"Reading LOB input failed");
					return;
				}

				stream.next_in = s_inbuf;
				stream.avail_in = n;

				inlob_remaining -= n;
				input_pos += n;
			}

			status = inflate(&stream, Z_SYNC_FLUSH);

			if ((status == Z_STREAM_END) || (!stream.avail_out))
			{
				// Output buffer is full, or decompression is done, so write buffer to output LOB
				uint n = BUF_SIZE - stream.avail_out;
				rc = sqludf_append(output,
								  (unsigned char *) s_outbuf, n, &length);
				if (rc || length != n)
				{
					strcpy(sqlstate,"38U04");
					strcpy(msgtext,"Writing LOB output failed");
					return;
				}
				stream.next_out = s_outbuf;
				stream.avail_out = BUF_SIZE;
			}

			if (status == Z_STREAM_END)
				break;
			else if (status != Z_OK)
			{
				strcpy(sqlstate,"38U05");
				strcpy(msgtext,"Inflate failed");
				return;
			}
		}

		if (inflateEnd(&stream) != Z_OK)
		{
				strcpy(sqlstate,"38U06");
				strcpy(msgtext,"Inflate End failed");
				return;
		}
	}
}